/*************************************************************
 *  AltGr key FIX for Windows 7 NTVDM
 *
 *  (C)oded by Dipl.-Ing. (FH) Ludwig Ertl, 2015
 *
 * This little VDD driver fixes a problem with the Windows 
 * Vista-10 NTVDM where the right AltGr key is incorretly 
 * passed to the NTVDM as left Ctrl+Right alt which leads to 
 * wrong keyboard input and i.e. prevents the german keyboard 
 * driver from working.
 * This is done via analyzing an parsing of keyboard input and
 * may lead to some delay (which hopefully shouldn't be 
 * noticable)
 *
 * Thank you Microsoft for not fixing this bug for over 8 years!
 *
 *************************************************************
 * Module:   altgrfix.c
 * Descr.:   Main module and dispatcher. 
 * License:  GPL 3
 * Date  :   03.12.2015
 * Changelog:
 *************************************************************/

// ------------------------- INCLUDES -------------------------
#define i386
#include <windows.h> 
#include <vddsvc.h>

// ------------------------ LIBRARIES -------------------------
#pragma comment (lib, "ntvdm.lib")

// ------------------------- DEFINES --------------------------
#ifdef _DEBUG
#define DBG_CALLS   // Debug calls on Debug-Console
#endif

typedef HANDLE (*fpGetConsoleInputWaitHandle)(VOID);
typedef BOOL (__stdcall* fpReadConsoleInputExW)(HANDLE,PINPUT_RECORD,DWORD,PDWORD,DWORD);

// ------------------------ VARIABLES -------------------------

// ------------------------ PROTOTYPES ------------------------
BOOL ReadConsoleInputExWHook(HANDLE hInput, PINPUT_RECORD pRecord, DWORD nItems, PDWORD pdwRead, DWORD dwFlags);
static BOOL Hook_IAT(LPBYTE hMod, char *pszLib, char *pszFunct, LPVOID NewFun) ;
static fpReadConsoleInputExW       ReadConsoleInputExW;
static fpGetConsoleInputWaitHandle GetConsoleInputWaitHandle;

// =============================================================
// Public (Exports)
// =============================================================
__declspec(dllexport) BOOL __cdecl VDDInitialize(IN PVOID DllHandle, IN ULONG Reason, 
		IN PCONTEXT Context OPTIONAL)
{
    switch (Reason) {

    case DLL_PROCESS_ATTACH:
	{
		HMODULE hNtvdm, hKernel32 = GetModuleHandle("KERNEL32.dll");

		GetConsoleInputWaitHandle = (fpGetConsoleInputWaitHandle)GetProcAddress(hKernel32, "GetConsoleInputWaitHandle");
		ReadConsoleInputExW = (fpReadConsoleInputExW)GetProcAddress(hKernel32, "ReadConsoleInputExW");
		if (!GetConsoleInputWaitHandle || !ReadConsoleInputExW)
		{
			OutputDebugString ("Kernel32 console functions not found");
			return FALSE;
		}

        /* Install hook */
		if (!(hNtvdm = GetModuleHandle("ntvdm.exe")))
		{
			OutputDebugString ("NTVDM.EXE Module not found in address space");
			return FALSE;
		}

        if (!Hook_IAT((PBYTE)hNtvdm, "kernel32.dll", "ReadConsoleInputExW", ReadConsoleInputExWHook))
        {
            OutputDebugString ("Hooking ReadConsoleInputExW failed");
			return FALSE;
        }
        OutputDebugString ("AltGrFix loaded successfully");
        break;
	}
    case DLL_PROCESS_DETACH:
        OutputDebugString ("AltGrFix unloaded");
        break;
    default:
        break;
    }
 
    return TRUE;
}

// =============================================================
// Private
// =============================================================
static BOOL Hook_IAT(LPBYTE hMod, char *pszLib, char *pszFunct, LPVOID NewFun) 
{
	PIMAGE_DOS_HEADER DosHeader = (PIMAGE_DOS_HEADER)hMod;
	PIMAGE_NT_HEADERS NtHeaders = (PIMAGE_NT_HEADERS)(hMod + DosHeader->e_lfanew);
	PIMAGE_IMPORT_DESCRIPTOR idata = (PIMAGE_IMPORT_DESCRIPTOR)(hMod + NtHeaders->OptionalHeader.DataDirectory[IMAGE_DIRECTORY_ENTRY_IMPORT].VirtualAddress);
	int i;
	char *pszDLL;

	for (; idata->Name; idata++)
	{
		pszDLL = (char*)(hMod + idata->Name);
		if (!lstrcmpi(pszDLL,pszLib)) 
		{
			PIMAGE_THUNK_DATA ThunkData = (PIMAGE_THUNK_DATA)(hMod + idata->OriginalFirstThunk);
			DWORD* Address = (DWORD*)(hMod + idata->FirstThunk);
			for (i = 0; ThunkData->u1.ForwarderString; i++)
			{
				if (!lstrcmpi((char*)(hMod + ThunkData->u1.ForwarderString + 2), pszFunct))
				{
					DWORD OldProt;
					VirtualProtect(&Address[i], 8, PAGE_READWRITE, &OldProt);
					Address[i] = (DWORD)NewFun;
					VirtualProtect(&Address[i], 8, OldProt, &OldProt);
					return TRUE;

				} ThunkData++;
			}
		}
	}
	return FALSE;
}

BOOL ReadConsoleInputExWHook(HANDLE hInput, PINPUT_RECORD pRecord, DWORD nItems, PDWORD pdwRead, DWORD dwFlags)
{
	BOOL bRet;
	DWORD i;
	static INPUT_RECORD LastRecord = {0};

	if (bRet = ReadConsoleInputExW(hInput, pRecord, nItems>1?nItems-1:nItems, pdwRead, dwFlags))
	{
		for (i=0; i<*pdwRead; i++)
		{
			if (pRecord[i].EventType == KEY_EVENT && pRecord[i].Event.KeyEvent.uChar.AsciiChar == 0 &&
				pRecord[i].Event.KeyEvent.bKeyDown)
			{
				if ((pRecord[i].Event.KeyEvent.dwControlKeyState & 0x1F) == LEFT_CTRL_PRESSED)
				{
					BOOL bNextOk = TRUE;

					if (i==*pdwRead-1)
					{
						DWORD dwRead;

						if (WaitForSingleObject(GetConsoleInputWaitHandle(), 10) == WAIT_OBJECT_0 &&
							(bNextOk = ReadConsoleInputExW(hInput, &pRecord[i+1], 1, &dwRead, dwFlags)))
							(*pdwRead)++;
					}
					if (bNextOk)
					{
						if (pRecord[i+1].EventType == KEY_EVENT && pRecord[i+1].Event.KeyEvent.bKeyDown &&
							pRecord[i+1].Event.KeyEvent.uChar.AsciiChar == 0 &&
							(pRecord[i+1].Event.KeyEvent.dwControlKeyState & RIGHT_ALT_PRESSED) == RIGHT_ALT_PRESSED)
						{
							(*pdwRead)--;
							RtlMoveMemory(&pRecord[i], &pRecord[i+1], (*pdwRead-i)*sizeof(INPUT_RECORD));
						}
					}
				}
				else if ((pRecord[i].Event.KeyEvent.dwControlKeyState & 
						(LEFT_CTRL_PRESSED | RIGHT_ALT_PRESSED | ENHANCED_KEY)) == 
						(LEFT_CTRL_PRESSED | RIGHT_ALT_PRESSED) && 
						(LastRecord.Event.KeyEvent.dwControlKeyState & 
						(LEFT_CTRL_PRESSED | RIGHT_ALT_PRESSED | ENHANCED_KEY)) == 
						(LEFT_CTRL_PRESSED | RIGHT_ALT_PRESSED | ENHANCED_KEY))
				{
					(*pdwRead)--;
					RtlMoveMemory(&pRecord[i], &pRecord[i+1], (*pdwRead-i)*sizeof(INPUT_RECORD));
					if (!*pdwRead) bRet = ReadConsoleInputExW(hInput, pRecord, nItems, pdwRead, dwFlags);
				}
				LastRecord = pRecord[i];
			}
		}
	}
	return bRet;
}


